r"""Simple demonstrations of manipulating Symbols and SymbolDict in inquanto"""

# imports
from sympy import Symbol, pi

from pytket import Circuit
from pytket.circuit import fresh_symbol
from inquanto.core import SymbolSet, SymbolDict
from inquanto.ansatzes import CircuitAnsatz

# We create a very simple SymbolDict
simple_sd = SymbolDict(a=1, b=2)
print(simple_sd)

# SymbolDicts contain sympy Symbols and coefficients
print(type(simple_sd.symbols[0]))

# you can copy SymbolDicts
second_simple = simple_sd.copy()

# and compare them
print("Example SymbolDicts are the same: " + str(second_simple == simple_sd))

# SymbolDicts can be printed to a table
print(second_simple.df())

# Symbols can be removed from a SymbolDict
second_simple.discard("b")
# and now the copy is no longer equal
print("Example SymbolDicts are the same: " + str(second_simple == simple_sd))


# Parameterised circuits are an important tool, and we often use them in ansatzes
# here we construct a pytket circuit with symbols and demonstrate their use
# in the ansatz class


# create a pytket circuit with symbols and then create classes from those symbols
circ = Circuit(2)
circ.Ry(-2 * fresh_symbol("a") / pi, 0)
circ.Rz(-2 * fresh_symbol("b") / pi, 1)
circ.Rx(-2 * fresh_symbol("c") / pi, 1)
circ.Ry(-2 * fresh_symbol("c") / pi, 0)

# create a SymbolDict for the example circuit
sd = SymbolDict.from_circuit(circ)

# print free symbols without values
print("SymbolDict with no coefficients")
print(sd)

# create a SymbolSet (no coefficients)
circ_ss = SymbolSet.from_circuit(circ)
print("SymbolSet")
print(circ_ss)


# we can combine SymbolDicts and Symbols using update
# create a second symbolset
ss2 = SymbolSet(["a", "d", "e"])
# we can extend symbol sets without repeats
circ_ss.update(ss2)
print("Extended set")
print(circ_ss)


# .update below does nothing as the copy contains only repeats
ss3 = circ_ss.copy()
circ_ss.update(ss3)
print(ss3)


# ansatz have methods for manipulating symbols
# we can create an inquanto ansatz which carries the circuit (and the symbols) from before
ansatz = CircuitAnsatz(circ)
print("ANSATZ CIRCUIT FREE SYMBOLS")
print(ansatz.free_symbols())

# we can copy and rename ansatz symbols
# this is useful when comparing similar states
ansatz1 = ansatz.copy()
# symbols_substitution here iterates over free_symbols and appends to their names
# symbol_substitution is a very flexible method
ansatz1.symbol_substitution(r"{}_X")
print("RENAMED ANSATZ CIRCUIT FREE SYMBOLS")
print(ansatz1.free_symbols())


# we can construct a SymbolDict for the ansatz by setting values to 0
example_parameters = ansatz.state_symbols.construct_zeros()
print("ZEROS")
print(example_parameters)

# SymbolDict.update will replace values or extend the dict
example_parameters.update(SymbolDict(a=5, b=1))
print(example_parameters)

# or randomly seeding values in the ansatz free symbols
example_parameters = ansatz.state_symbols.construct_random(seed=0, mu=0.0, sigma=0.3)
print("RANDOM")
print(example_parameters)

# we can obtain and manipulate coefficients and input them into a new dict
new_array = []
constant = 1.2
for coeff in example_parameters.to_array():
    new_array.append(coeff + constant)
print("FROM ARRAY SHIFTED RANDOMS")
example_parameters = ansatz.state_symbols.construct_from_array(new_array)
print(example_parameters)

# or from a given manually input array
example_parameters = ansatz.state_symbols.construct_from_array([0, 1.4, 0.0003, 4.368])
print("FROM ARRAY MANUALLY DEFINED")
print(example_parameters)


extra_symbols = SymbolDict({"x2": 0.01, "k_k": 5.2})
print("EXTENDED DICT")
print(example_parameters.update(extra_symbols))
# this symboldict is overcomplete, and the additional symbols will be ignored when substituting

# symbol_substitution can be used to insert numerics into free_symbols in ansatz
ansatz.symbol_substitution(example_parameters)
# After complete symbol substitution there are no free_symbols
# this is not reversible
print(ansatz.free_symbols())
